#include <stack>

#include "compact_interval_stabbing_new.h"

compact_interval_stabbing_new::compact_interval_stabbing_new(interval *intervals, uint size, uint lowerBlockSize, uint upperBlockSize, char *encoder, int scaleFactor) {
  n = size+1;
  //+1 because we use a fictitious interval to generalize the transformation of the
  //queries to the rank space
  double *coordsL = new double[n-1];
  double *coordsU = new double[n-1];
  int coordsLPos = 0;int coordsUPos = 0;
  
  ids = new long[size];
  
  coord_interval *coordInts = new coord_interval[2 * size];
  
  for(uint i = 0; i < size; i++) {
    coord_interval ci = {intervals[i].getMin(), true, &intervals[i]};
    coord_interval ci1 = {intervals[i].getMax(), false, &intervals[i]};
    coordInts[2*i] = ci;
    coordInts[2*i+1] = ci1;
  }
  
  qsort(coordInts, 2*size, sizeof(coord_interval), cmpCoordInterval);
  
  uint idsSize = 0;
  uint auxTreeBitmapSize = (2*n+W-1)/W;
  uint auxTreeBitmap[auxTreeBitmapSize];
  uint auxTreeBitmapCPos = 0;
  for(uint i = 0; i < auxTreeBitmapSize; i++) {
    auxTreeBitmap[i] = 0;
  }
  bitput(auxTreeBitmap, auxTreeBitmapCPos++, 1, 0);//tree root
  stack<double> auxStack;
  for(uint i = 0; i < 2*size; i++) {
    //cout << coordInts[i].coord << " (" << coordInts[i].intervalPtr->getId() << ") ";
    if(coordInts[i].min) {
      coordsL[coordsLPos++] = coordInts[i].coord;
      coordsU[coordsUPos++] = coordInts[i].intervalPtr->getMax();
      ids[idsSize++] = coordInts[i].intervalPtr->getId();
      while(!auxStack.empty() && auxStack.top() < coordInts[i].intervalPtr->getMax()) {
	bitput(auxTreeBitmap, auxTreeBitmapCPos++, 1, 1);
	auxStack.pop();
      }
      bitput(auxTreeBitmap, auxTreeBitmapCPos++, 1, 0);
      auxStack.push(coordInts[i].intervalPtr->getMax());
    } else {
      while(!auxStack.empty() && auxStack.top() <= coordInts[i].intervalPtr->getMax()) {
	bitput(auxTreeBitmap, auxTreeBitmapCPos++, 1, 1);
	auxStack.pop();
      }
    }
  }
  //fictitious interval begin
  bitput(auxTreeBitmap, auxTreeBitmapCPos++, 1, 0);
  bitput(auxTreeBitmap, auxTreeBitmapCPos++, 1, 1);
  //fictitious interval end
  bitput(auxTreeBitmap, auxTreeBitmapCPos++, 1, 1);//close tree root
  
  delete []coordInts;
  
  tree = new tree_ff(auxTreeBitmap, 2*(n+1), OPT_FAST_PREORDER_SELECT);
  
  //upperCoords =  new UpperCoordinates(coordsU, n-1, scaleFactor);
  if(!strcmp(encoder,PLAIN)) {
    lowerCoords = new PlainArray(coordsL, n-1);
    upperCoords =  new PlainArray(coordsU, n-1);
  } else {
    lowerCoords =  new EncodedArray(coordsL, n-1, lowerBlockSize, encoder, scaleFactor);
    upperCoords =  new EncodedArrayXOR(coordsU, n-1, upperBlockSize, encoder, scaleFactor);
  }
  delete []coordsL;
  delete []coordsU;
  
  nodesDeque = new uint[n];
}

compact_interval_stabbing_new::~compact_interval_stabbing_new() {
  delete []nodesDeque;
  delete tree;
  delete []ids;
  delete lowerCoords;
  delete upperCoords;
}

int compact_interval_stabbing_new::queryMark(interval *queryInterval, long *result, uint *resultSize, uint *intersector) {
  ndHead = 0;
  double qiMin = queryInterval->getMin();
  double qiMax = queryInterval->getMax();
  if(qiMax < lowerCoords->access(0)) return 0;
    
    int cNodePreorder = lowerCoords->searchRight(qiMax)+1;
    int cNodePos = tree->Preorden_Select(cNodePreorder+1);
    nodesDeque[ndHead++] = cNodePos;
    
    int tmpCNodePos = tree->Parent(cNodePos);
    while(tmpCNodePos != 0) {
      cNodePreorder = tree->Preorder_Rank(tmpCNodePos);
      bitset(intersector, ids[cNodePreorder-2]);//-2 because the root is not represented
      nodesDeque[ndHead++] = tmpCNodePos;
      tmpCNodePos = tree->Parent(tmpCNodePos);
    }
    while(ndHead > 0) {
      cNodePos = tree->Prev_Sibling(nodesDeque[--ndHead]);
      while(cNodePos != -1) {
	cNodePreorder = tree->Preorder_Rank(cNodePos);
	if(upperCoords->access(cNodePreorder-2) < qiMin) {
	  break;
	}
	bitset(intersector, ids[cNodePreorder-2]);
	nodesDeque[ndHead++] = cNodePos;
	cNodePos = tree->Last_Child(cNodePos);
      }
    }
    
    return 0;
}

int compact_interval_stabbing_new::queryReport(interval *queryInterval, long *result, uint *resultSize, uint *intersector) {
  *resultSize = 0;
  ndHead = 0;
  double qiMin = queryInterval->getMin();
  double qiMax = queryInterval->getMax();
  if(qiMax < lowerCoords->access(0)) return 0;
    
    int cNodePreorder = lowerCoords->searchRight(qiMax)+1;
    int cNodePos = tree->Preorden_Select(cNodePreorder+1);
    nodesDeque[ndHead++] = cNodePos;
    
    int tmpCNodePos = tree->Parent(cNodePos);
    while(tmpCNodePos != 0) {
      cNodePreorder = tree->Preorder_Rank(tmpCNodePos);
      if(bitget(intersector, ids[cNodePreorder-2])) {
	result[(*resultSize)++] = ids[cNodePreorder - 2];
      }
      nodesDeque[ndHead++] = tmpCNodePos;
      tmpCNodePos = tree->Parent(tmpCNodePos);
    }
    while(ndHead > 0) {
      cNodePos = tree->Prev_Sibling(nodesDeque[--ndHead]);
      while(cNodePos != -1) {
	cNodePreorder = tree->Preorder_Rank(cNodePos);
	if(upperCoords->access(cNodePreorder-2) < qiMin) {
	  break;
	}
	if(bitget(intersector, ids[cNodePreorder-2])) {
	  result[(*resultSize)++] = ids[cNodePreorder - 2];
	}
	nodesDeque[ndHead++] = cNodePos;
	cNodePos = tree->Last_Child(cNodePos);
      }
    }
    
    return 0;
}

uint compact_interval_stabbing_new::memoryUsage() {
  //std::cout << lowerCoords->memoryUsage()+upperCoords->memoryUsage() << ";" << n * sizeof(uint) << ";" << tree->size()+sizeof(compact_interval_stabbing_new) << endl;
  return lowerCoords->memoryUsage()
  + upperCoords->memoryUsage()
  + n * sizeof(uint) //ids
  + tree->size()
  + sizeof(compact_interval_stabbing_new);
}
