#include <stack>
#include <deque>

#include "isaac_interval_stabbing.h"

isaac_interval_stabbing::isaac_interval_stabbing(interval *intervals, uint size) {
  n = size;
  coords = new double[n];
  
  coord_interval *coordInts = new coord_interval[2 * n];
  
  for(int i = 0; i < n; i++) {
    coord_interval ci = {intervals[i].getMin(), true, &intervals[i]};
    coord_interval ci1 = {intervals[i].getMax(), false, &intervals[i]};
    coordInts[2*i] = ci;
    coordInts[2*i+1] = ci1;
  }
  
  qsort(coordInts, 2*n, sizeof(coord_interval), cmpCoordInterval);
  
  uint coordsPos = 0;
  for(int i = 0; i < 2*n; i++) {
    if(coordInts[i].min) {
      coords[coordsPos++] = coordInts[i].coord;
      coordInts[i].intervalPtr->setMinRank(i);
    } else {
      coordInts[i].intervalPtr->setMaxRank(i);
    }
  }
  
  //create the tree and start
  uint startPos = 0;
  start = new isaac_node*[n+1];
  root = new isaac_node;
  root->parent = root->rightchild = root->leftsibling = NULL;
  root->coord = coordInts[2*n-1].intervalPtr->getMax()+1;
  root->id = -1;
  stack<isaac_node*> nodesStack;
  stack<interval*> intervalsStack;
  nodesStack.push(root);
  isaac_node *tmpNode, *tmpParent;
  for(int i = 0; i < 2*n; i++) {
    if(coordInts[i].min) {
      tmpNode = new isaac_node;
      tmpNode->id = coordInts[i].intervalPtr->getId();
      tmpNode->coord = coordInts[i].intervalPtr->getMax();
      tmpNode->rightchild = NULL;
      while(!intervalsStack.empty() && intervalsStack.top()->getMaxRank() < i) {
	intervalsStack.pop();
	nodesStack.pop();
      }
      if(nodesStack.empty()) {
	tmpParent = root;
      } else {
	tmpParent = nodesStack.top();
	while(tmpParent->coord < coordInts[i].intervalPtr->getMax()) {
	  tmpParent = tmpParent->parent;
	}
      }
      tmpNode->leftsibling = tmpParent->rightchild;
      tmpNode->parent = tmpParent;
      tmpParent->rightchild = tmpNode;
      nodesStack.push(tmpNode);
      intervalsStack.push(coordInts[i].intervalPtr);
      start[startPos++] = tmpNode;
    } else {
      while(!intervalsStack.empty() && intervalsStack.top()->getMaxRank() <= i) {
	nodesStack.pop();
	intervalsStack.pop();
      }
    }
  }
  
  isaac_node *fictNode = new isaac_node;
  tmpNode = root->rightchild;
  fictNode->id = -1;
  fictNode->parent = root;
  fictNode->coord = tmpNode->coord+1;
  fictNode->rightchild = NULL;
  fictNode->leftsibling = tmpNode;
  root->rightchild = fictNode;
  start[n] = fictNode;
  
  delete []coordInts;
  
  nodesDeque = new isaac_node*[n];
}

void recursive_free(isaac_node *node) {
  if(node->rightchild != NULL) {
    recursive_free(node->rightchild);
  }
  if(node->leftsibling != NULL) {
    recursive_free(node->leftsibling);
  }
  delete node;
}

isaac_interval_stabbing::~isaac_interval_stabbing() {
  recursive_free(root);
  delete []start;
  delete []coords;
  delete []nodesDeque;
}

/*
int isaac_interval_stabbing::query(const double queryPoint, long *result, uint *resultSize) {
  *resultSize = 0;
  int pos;
  bool found;
  found = rightBinarySearch(queryPoint, coords, 2*n, &pos);
  //cout << queryPoint << " " << found << " " << pos << "\n";
  
  pos = 2*pos;
  if(!found) {
    if(pos < 0) {return 0;}
    pos++;
    }
    
    isaac_node *cNode;
    deque<isaac_node*> nodesDeque;
    for(cNode = start[pos]; cNode->parent != NULL; cNode = cNode->parent) {
      nodesDeque.push_front(cNode);
    }
    while(!nodesDeque.empty()) {
      cNode = nodesDeque.back();
      nodesDeque.pop_back();
      result[(*resultSize)++] = cNode->id;
      
      cNode = cNode->leftsibling;
      while(cNode != NULL && *cNode->coord >= queryPoint) {
	nodesDeque.push_back(cNode);
	cNode = cNode->rightchild;
      }
    }
    
    return 0;
    }
    */

int isaac_interval_stabbing::queryMark(interval *queryInterval, long *result, uint *resultSize, uint *intersector) {
  ndHead = 0;
  int pos;
  double qiMin = queryInterval->getMin();
  double qiMax = queryInterval->getMax();
  rightBinarySearch(qiMax, coords, n, &pos);
  if(pos < 0) {return 0;}
  
  isaac_node *cNode = start[pos+1];
  nodesDeque[ndHead++] = cNode;
  cNode = cNode->parent;
  while(cNode->parent != NULL) {
    nodesDeque[ndHead++] = cNode;
    bitset(intersector, cNode->id);
    cNode = cNode->parent;
  }
  while(ndHead > 0) {
    cNode = nodesDeque[--ndHead]->leftsibling;    
    while(cNode != NULL && cNode->coord >= qiMin) {
      bitset(intersector, cNode->id);
      nodesDeque[ndHead++] = cNode;
      cNode = cNode->rightchild;
    }
  }
  
  return 0;
}

int isaac_interval_stabbing::queryReport(interval *queryInterval, long *result, uint *resultSize, uint *intersector) {
  (*resultSize) = 0;
  ndHead = 0;
  int pos;
  double qiMin = queryInterval->getMin();
  double qiMax = queryInterval->getMax();
  rightBinarySearch(qiMax, coords, n, &pos);
  if(pos < 0) {return 0;}
  
  isaac_node *cNode = start[pos+1];
  nodesDeque[ndHead++] = cNode;
  cNode = cNode->parent;
  while(cNode->parent != NULL) {
    if(bitget(intersector, cNode->id))
      result[(*resultSize)++] = cNode->id;
    nodesDeque[ndHead++] = cNode;
    cNode = cNode->parent;
  }
  while(ndHead > 0) {
    cNode = nodesDeque[--ndHead]->leftsibling;    
    while(cNode != NULL && cNode->coord >= qiMin) {
      if(bitget(intersector, cNode->id))
	result[(*resultSize)++] = cNode->id;
      nodesDeque[ndHead++] = cNode;
      cNode = cNode->rightchild;
    }
  }


  return 0;
}

uint isaac_interval_stabbing::memoryUsage() {
  return n * sizeof(float)//although in the code we use doubles
  + (n+1) * sizeof(isaac_node)
  +  (n+1) * sizeof(isaac_node*) //start
  + sizeof(isaac_interval_stabbing);
}
