#include <iostream>
#include <cstring>
#include <cstdlib>

#ifndef TIEMPO_H
#include "../util/tiempo.h"
#endif

#include "mbr_range.h"

int cmplong (const void * a, const void * b)
{
  return ( *(long*)a - *(long*)b );
}

int main(int argc, char** argv)
{
	if (argc < 12)
	{
		printf("Usage: %s input_file file_type outputFile queryFiles query_files_type [COMPACT|NOCOMPACT] encoder lowerBlockSize upperBlockSize [scaleFactor|1] [test|time].\n", argv[0]);
		return -1;
	}

	uint fileType = 0;
	uint queryFileType = 0;
	uint structureType = 0;

	if (strcmp(argv[2], "IdX1Y1X2Y2") == 0) fileType = 0;
	else if (strcmp(argv[2], "X1Y1X2Y2") == 0) fileType = 1;
	else if (strcmp(argv[2], "IdY1X1Y2X2") == 0) fileType = 2;
	else if (strcmp(argv[2], "Y1X1Y2X2") == 0) fileType = 3;
	else
	{
		printf("Unknown file type.");
		return -1;
	}
	if (strcmp(argv[argc - 7], "IdX1Y1X2Y2") == 0) queryFileType = 0;
	else if (strcmp(argv[argc - 7], "X1Y1X2Y2") == 0) queryFileType = 1;
	else if (strcmp(argv[argc - 7], "IdY1X1Y2X2") == 0) queryFileType = 2;
	else if (strcmp(argv[argc - 7], "Y1X1Y2X2") == 0) queryFileType = 3;
	else
	{
		printf("Unknown query file type.");
		return -1;
	}
	if (strcmp(argv[argc - 6], "COMPACT") == 0) structureType = COMPACT;
	else if (strcmp(argv[argc - 6], "COMPACTNEW") == 0) structureType = COMPACT_NEW;
	else if (strcmp(argv[argc - 6], "NOCOMPACT") == 0) structureType = NO_COMPACT;
	else if (strcmp(argv[argc - 6], "MAXIMAL") == 0) structureType = MAXIMAL;
	else
	{
		printf("Unknown structure type.");
		return -1;
	}

	FILE* outputFile = fopen(argv[3], "a");

	//read input elements
	uint number_of_input_bboxes;
	mbr *input_bboxes;
	if(fileType == 0) {
		mbr::readMBRs(argv[1], &input_bboxes, &number_of_input_bboxes);
	} else if(fileType == 1) {
		mbr::readMBRsAndAssignIDs(argv[1], &input_bboxes, &number_of_input_bboxes);
	} else {
		printf("Not supported file type.");
	}

	//printf("Creando estructuras..\n");
	mbr_range is = mbr_range(input_bboxes, number_of_input_bboxes, structureType, atoi(argv[argc - 4]), atoi(argv[argc - 3]), argv[argc - 5], atoi(argv[argc - 2]));
	//printf("Estructuras creadas..\n");
	//printf("%u;\n", is.memoryUsage());


	long* resultWT = (long*) malloc(number_of_input_bboxes * sizeof(long));
	int k;
	for(k = 4; k < argc - 7; k++) {
		//read input queries
		uint number_of_input_queries;
		mbr *input_queries;
		if(queryFileType == 0) {
			mbr::readMBRs(argv[k], &input_queries, &number_of_input_queries);
		} else if(queryFileType == 1) {
			mbr::readMBRsAndAssignIDs(argv[k], &input_queries, &number_of_input_queries);
		} else {
		  printf("Not supported query type.");
		}
		
		uint i,j;
		uint resultSizeWT;
		double time = 0;
		
		if(!strcmp(argv[argc - 1], "time")) {
			for(i = 0; i < number_of_input_queries; i++) {
				tTiempo mitiempo1;
				tTiempo mitiempo2;
				getTiempo(&mitiempo1);
				
				for(j = 0; j < 5; j++) {
					resultSizeWT = 0;
					is.query(&input_queries[i], resultWT, &resultSizeWT);
				}
				
				getTiempo(&mitiempo2);
				time += ((mitiempo2.user - mitiempo1.user) / 5);
			}
			fprintf(outputFile, "%lf;", time / number_of_input_queries);
		} else if(!strcmp(argv[argc - 1], "test")) {
			uint error = 0;
			for(i = 0; i < number_of_input_queries; i++) {
				resultSizeWT = 0;
				is.query(&input_queries[i], resultWT, &resultSizeWT);
				printf("Query %ld: %d elements: ", input_queries[i].getId(), (uint)resultSizeWT);
				qsort(resultWT, resultSizeWT, sizeof(long), cmplong);
				for(j = 0; j < resultSizeWT; j++) {
					printf("%ld; ", resultWT[j]);
				}
				printf("\n");
				
			}
			if(!error) {
				fprintf(outputFile, "All tests OK.\n");
			}
		} else {
			printf("Unknown option: %s\n", argv[argc - 1]);
		}
	
		free(input_queries);
	}
	
	free(resultWT);
	free(input_bboxes);
	
	if(!strcmp(argv[argc - 1], "time")) {
	  fprintf(outputFile, "%u;", is.memoryUsage());
	}
	fprintf(outputFile, "\n");
	fclose(outputFile);

	return 0;
}
