#include "maximal_interval_stabbing.h"

maximal_interval_stabbing::maximal_interval_stabbing(interval *intervals, uint size, uint lowerBlockSize, uint upperBlockSize, char *encoder, int scaleFactor) {
  n = size;
  
  //Maximal decomposition
  numMaxSets = 1;
  uint *maximalSeq = new uint[n];
  double *maximalEnds = new double[n];
  uint *numElementsInSet = new uint[n];
  qsort(intervals, n, sizeof(interval), cmpInterval);
  maximalSeq[0] = 0;
  maximalEnds[0] = intervals[0].getMax();
  numElementsInSet[0] = 1;
  for(int i = 1; i < n; i++) {
    bool found = false;
    for(int j = 0; j < numMaxSets; j++) { //we could use binary search
      if(intervals[i].getMax() >= maximalEnds[j]) {
	found = true;
	maximalSeq[i] = j;
	maximalEnds[j] = intervals[i].getMax();
	numElementsInSet[j] = numElementsInSet[j]+1;
	break;
      }
    }
    if(!found) {
      maximalSeq[i] = numMaxSets;
      maximalEnds[numMaxSets] = intervals[i].getMax();
      numElementsInSet[numMaxSets] = 1;
      numMaxSets++;
    }
  }
  delete [] maximalEnds;
  //End maximal decomposition
  
  double **upCoordsD = new double*[numMaxSets];
  double **lowCoordsD = new double*[numMaxSets];
  lowerCoords = new CoordinatesStorage*[numMaxSets];
  upperCoords = new CoordinatesStorage*[numMaxSets];
  ids = new long*[numMaxSets];
  for(int i = 0; i < numMaxSets; i++) {
    lowCoordsD[i] = new double[numElementsInSet[i]];
    upCoordsD[i] = new double[numElementsInSet[i]];
    ids[i] = new long[numElementsInSet[i]];
  }
  
  uint *cCountersInSet = new uint[numMaxSets];
  for(int i = 0; i < numMaxSets; i++) cCountersInSet[i] = 0;
  
  for(int i = 0; i < n; i++) {
    uint cSet = maximalSeq[i];
    uint cPos = cCountersInSet[cSet];
    lowCoordsD[cSet][cPos] = intervals[i].getMin();
    upCoordsD[cSet][cPos] = intervals[i].getMax();
    ids[cSet][cPos] = intervals[i].getId();    
    cCountersInSet[cSet] = cCountersInSet[cSet]+1;
  }
  delete [] cCountersInSet;
  
  for(int i = 0; i < numMaxSets; i++) {
    if(!strcmp(encoder,PLAIN)) {
      lowerCoords[i] = new PlainArray(lowCoordsD[i], numElementsInSet[i]);
      upperCoords[i] =  new PlainArray(upCoordsD[i], numElementsInSet[i]);
    } else {
      lowerCoords[i] =  new EncodedArray(lowCoordsD[i], numElementsInSet[i], lowerBlockSize, encoder, scaleFactor);
      upperCoords[i] =  new EncodedArray(upCoordsD[i], numElementsInSet[i], upperBlockSize, encoder, scaleFactor);
    }
    delete lowCoordsD[i];
    delete upCoordsD[i];
  }  
  delete [] numElementsInSet;
  delete [] upCoordsD;
  delete [] lowCoordsD;
  
  delete [] maximalSeq;
}

maximal_interval_stabbing::~maximal_interval_stabbing() {
    for(int i = 0; i < numMaxSets; i++) {
      delete lowerCoords[i];
      delete upperCoords[i];
      delete [] ids[i];
    }
    delete [] lowerCoords;
    delete [] upperCoords;
    delete [] ids;
}


int maximal_interval_stabbing::queryMark(interval *queryInterval, long *result, uint *resultSize, uint *intersector) {
  int pos;
  double qiMin = queryInterval->getMin();
  double qiMax = queryInterval->getMax();
  double cValue;
  uint binPos;
  for(int i = 0; i < numMaxSets; i++) {
    pos = upperCoords[i]->searchLeft(qiMin)-1;
    if(pos >= lowerCoords[i]->getListLength()) continue;
    cValue = lowerCoords[i]->accessSeq(pos, true, 0, 0, &binPos);
    while(cValue-0.00001 <= qiMax) {
      bitset(intersector, ids[i][pos]);
      pos++;
      if(pos >= lowerCoords[i]->getListLength()) break;
      cValue = lowerCoords[i]->accessSeq(pos, false, binPos, cValue, &binPos);
    }
  }
  
  return 0;
}

int maximal_interval_stabbing::queryReport(interval *queryInterval, long *result, uint *resultSize, uint *intersector) {
  (*resultSize) = 0;
  
  int pos;
  double qiMin = queryInterval->getMin();
  double qiMax = queryInterval->getMax();
  double cValue;
  uint binPos;
  for(int i = 0; i < numMaxSets; i++) {
    pos = upperCoords[i]->searchLeft(qiMin)-1;
    if(pos >= lowerCoords[i]->getListLength()) continue;
    cValue = lowerCoords[i]->accessSeq(pos, true, 0, 0, &binPos);
    while(cValue-0.00001 <= qiMax) {
      if(bitget(intersector, ids[i][pos]))
	result[(*resultSize)++] = ids[i][pos];
      pos++;
      if(pos >= lowerCoords[i]->getListLength()) break;
      cValue = lowerCoords[i]->accessSeq(pos, false, binPos, cValue, &binPos);
    }
  }
  
  return 0;
}

uint maximal_interval_stabbing::memoryUsage() {
  uint coordsSize = 0;
  for(int i = 0; i < numMaxSets; i++) {
    coordsSize += lowerCoords[i]->memoryUsage();
    coordsSize += upperCoords[i]->memoryUsage();
  }
  //std::cout << coordsSize << ";" << n * sizeof(uint) << ";" << 3 * numMaxSets * sizeof(uint*) + sizeof(maximal_interval_stabbing) << endl;
  return coordsSize
  + n * sizeof(uint)//ids
  + 3 * numMaxSets * sizeof(uint*) //aditional pointers
  + sizeof(maximal_interval_stabbing);
}
