#include "maximal_interval_stabbing_new.h"

#ifndef TIEMPO_H
#include "../util/tiempo.h"
#endif

maximal_interval_stabbing_new::maximal_interval_stabbing_new(interval *intervals, uint size, uint lowerBlockSize, uint upperBlockSize, char *encoder, int scaleFactor) {
  n = size;
  
  //Maximal decomposition
  numMaxSets = 1;
  COORD *upCoordsTMP = new COORD[n];
  uint *maximalSeq = new uint[n];
  double *maximalEnds = new double[n];
  uint *numElementsInSet = new uint[n];
  qsort(intervals, n, sizeof(interval), cmpInterval);
  maximalSeq[0] = 0;
  maximalEnds[0] = intervals[0].getMax();
  upCoordsTMP[0].coord = intervals[0].getMax(); upCoordsTMP[0].id = 0;
  numElementsInSet[0] = 1;
  for(int i = 1; i < n; i++) {
    bool found = false;
    for(int j = 0; j < numMaxSets; j++) { //we could use binary search
      if(intervals[i].getMax() >= maximalEnds[j]) {
	found = true;
	maximalSeq[i] = j;
	maximalEnds[j] = intervals[i].getMax();
	numElementsInSet[j] = numElementsInSet[j]+1;
	upCoordsTMP[i].coord = intervals[i].getMax(); upCoordsTMP[i].id = j;
	break;
      }
    }
    if(!found) {
      maximalSeq[i] = numMaxSets;
      maximalEnds[numMaxSets] = intervals[i].getMax();
      numElementsInSet[numMaxSets] = 1;
      upCoordsTMP[i].coord = intervals[i].getMax(); upCoordsTMP[i].id = numMaxSets;
      numMaxSets++;
    }
  }
  delete [] maximalEnds;
  //End maximal decomposition
  
  double **upCoordsD = new double*[numMaxSets];
  double **lowCoordsD = new double*[numMaxSets];
  lowerCoords = new CoordinatesStorage*[numMaxSets];
  ids = new long*[numMaxSets];
  for(int i = 0; i < numMaxSets; i++) {
    lowCoordsD[i] = new double[numElementsInSet[i]];
    upCoordsD[i] = new double[numElementsInSet[i]];
    ids[i] = new long[numElementsInSet[i]];
  }
  
  uint *cCountersInSet = new uint[numMaxSets];
  for(int i = 0; i < numMaxSets; i++) cCountersInSet[i] = 0;
  
  for(int i = 0; i < n; i++) {
    uint cSet = maximalSeq[i];
    uint cPos = cCountersInSet[cSet];
    lowCoordsD[cSet][cPos] = intervals[i].getMin();
    upCoordsD[cSet][cPos] = intervals[i].getMax();
    ids[cSet][cPos] = intervals[i].getId();    
    cCountersInSet[cSet] = cCountersInSet[cSet]+1;
  }
  delete [] cCountersInSet;
  
  for(int i = 0; i < numMaxSets; i++) {
    if(!strcmp(encoder,PLAIN)) {
      lowerCoords[i] = new PlainArray(lowCoordsD[i], numElementsInSet[i]);
      //upperCoords[i] =  new PlainArray(upCoordsD[i], numElementsInSet[i]);
    } else {
      lowerCoords[i] =  new EncodedArray(lowCoordsD[i], numElementsInSet[i], lowerBlockSize, encoder, scaleFactor);
      //upperCoords[i] =  new EncodedArray(upCoordsD[i], numElementsInSet[i], upperBlockSize, encoder, scaleFactor);
    }
    delete lowCoordsD[i];
    delete upCoordsD[i];
  }  
  delete [] numElementsInSet;
  delete [] upCoordsD;
  delete [] lowCoordsD;

  double *tmpCoords = new double[n];
  qsort(upCoordsTMP, n, sizeof(COORD), cmpcoords);  
  for(int i = 0; i < n; i++) {
    tmpCoords[i] = upCoordsTMP[i].coord;
    maximalSeq[i] = upCoordsTMP[i].id;
  }
  delete [] upCoordsTMP;
  
  if(!strcmp(encoder,PLAIN)) {
    upperCoords =  new PlainArray(tmpCoords, n);
  } else {
    upperCoords =  new EncodedArray(tmpCoords, n, upperBlockSize, encoder, scaleFactor);
  }
  delete [] tmpCoords;

  cds_static::MapperNone * map = new cds_static::MapperNone();
  cds_static::wt_coder_huff *wc = new cds_static::wt_coder_huff(maximalSeq, n, map);
  maximalSeqWT = new cds_static::WaveletTree(maximalSeq, n, wc, new cds_static::BitSequenceBuilderRG(20), map);
  delete [] maximalSeq;
}

maximal_interval_stabbing_new::~maximal_interval_stabbing_new() {
    for(int i = 0; i < numMaxSets; i++) {
      delete lowerCoords[i];
      delete [] ids[i];
    }
    delete [] lowerCoords;
    delete upperCoords;
    delete [] ids;
    delete maximalSeqWT;    
}


int maximal_interval_stabbing_new::queryMark(interval *queryInterval, long *result, uint *resultSize, uint *intersector) {
  int pos;
  uint *minResult;
  double qiMin = queryInterval->getMin();
  double qiMax = queryInterval->getMax();
  
  pos = upperCoords->searchLeft(qiMin);
  maximalSeqWT->rankAll(pos-2, &minResult, numMaxSets);

  double cValue;
  uint binPos;
  for(int i = 0; i < numMaxSets; i++) {
    pos = minResult[i];
    if(pos >= lowerCoords[i]->getListLength()) continue;
    cValue = lowerCoords[i]->accessSeq(pos, true, 0, 0, &binPos);
    while(cValue-0.00001 <= qiMax) {
      bitset(intersector, ids[i][pos]);
      pos++;
      if(pos >= lowerCoords[i]->getListLength()) break;
      cValue = lowerCoords[i]->accessSeq(pos, false, binPos, cValue, &binPos);
    }
  }
  
  delete [] minResult;
  
  return 0;
}

int maximal_interval_stabbing_new::queryReport(interval *queryInterval, long *result, uint *resultSize, uint *intersector) {
  (*resultSize) = 0;
  
  int pos;
  uint *minResult;
  double qiMin = queryInterval->getMin();
  double qiMax = queryInterval->getMax();
  pos = upperCoords->searchLeft(qiMin);
  maximalSeqWT->rankAll(pos-2, &minResult, numMaxSets);

  double cValue;
  uint binPos;
  for(int i = 0; i < numMaxSets; i++) {
    pos = minResult[i];
    if(pos >= lowerCoords[i]->getListLength()) continue;
    cValue = lowerCoords[i]->accessSeq(pos, true, 0, 0, &binPos);
    while(cValue-0.00001 <= qiMax) {
      if(bitget(intersector, ids[i][pos]))
	result[(*resultSize)++] = ids[i][pos];
      pos++;
      if(pos >= lowerCoords[i]->getListLength()) break;
      cValue = lowerCoords[i]->accessSeq(pos, false, binPos, cValue, &binPos);
    }
  }
  
  return 0;
}

uint maximal_interval_stabbing_new::memoryUsage() {
  uint coordsSize = upperCoords->memoryUsage();
  for(int i = 0; i < numMaxSets; i++) {
    coordsSize += lowerCoords[i]->memoryUsage();
  }
  return coordsSize
  + n * sizeof(uint)//ids
  + 2 * numMaxSets * sizeof(uint*) //aditional pointers
  + maximalSeqWT->getSize()
  + sizeof(maximal_interval_stabbing_new);
}


void cleanIntersector(uint *intersector, uint intersectorSize) {
  for(int i = 0; i < intersectorSize; i++) {
    intersector[i] = 0;
  }
}

int query(mbr *queryMBR, long *result, uint *resultSize, uint *intersector, uint intersectorSize, maximal_interval_stabbing_new *x_intervals, maximal_interval_stabbing_new *y_intervals) {
  cleanIntersector(intersector, intersectorSize);
  long id = queryMBR->getId();
  double minx = queryMBR->getMinX();
  double maxx = queryMBR->getMaxX();
  double miny = queryMBR->getMinY();
  double maxy = queryMBR->getMaxY();
  interval x_interval = interval(id, minx, maxx);
  interval y_interval = interval(id, miny, maxy);
  
  //x_intervals->query(&x_interval, result, resultSize, intersector, QUERY_TYPE_MARK);
  //y_intervals->query(&y_interval, result, resultSize, intersector, QUERY_TYPE_CHECK_REPORT);
  x_intervals->queryMark(&x_interval, result, resultSize, intersector);
  y_intervals->queryReport(&y_interval, result, resultSize, intersector);
}

int cmplong (const void * a, const void * b)
{
  return ( *(long*)a - *(long*)b );
}

int main(int argc, char** argv)
{
  if (argc < 12)
  {
    printf("Usage: %s input_file file_type outputFile queryFiles query_files_type [COMPACT|NOCOMPACT] encoder lowerBlockSize upperBlockSize [scaleFactor|1] [test|time].\n", argv[0]);
    return -1;
  }
  
  uint fileType = 0;
  uint queryFileType = 0;
  uint structureType = 0;
  
  if (strcmp(argv[2], "IdX1Y1X2Y2") == 0) fileType = 0;
  else if (strcmp(argv[2], "X1Y1X2Y2") == 0) fileType = 1;
  else if (strcmp(argv[2], "IdY1X1Y2X2") == 0) fileType = 2;
  else if (strcmp(argv[2], "Y1X1Y2X2") == 0) fileType = 3;
  else
  {
    printf("Unknown file type.");
    return -1;
  }
  if (strcmp(argv[argc - 7], "IdX1Y1X2Y2") == 0) queryFileType = 0;
  else if (strcmp(argv[argc - 7], "X1Y1X2Y2") == 0) queryFileType = 1;
  else if (strcmp(argv[argc - 7], "IdY1X1Y2X2") == 0) queryFileType = 2;
  else if (strcmp(argv[argc - 7], "Y1X1Y2X2") == 0) queryFileType = 3;
  else
  {
    printf("Unknown query file type.");
    return -1;
  }
  
  FILE* outputFile = fopen(argv[3], "a");
  
  //read input elements
  uint number_of_input_bboxes;
  mbr *input_bboxes;
  if(fileType == 0) {
    mbr::readMBRs(argv[1], &input_bboxes, &number_of_input_bboxes);
  } else if(fileType == 1) {
    mbr::readMBRsAndAssignIDs(argv[1], &input_bboxes, &number_of_input_bboxes);
  } else {
    printf("Not supported file type.");
  }
  
  //printf("Creando estructuras..\n");
  uint intersectorSize = number_of_input_bboxes/W+1;
  uint *intersector = new uint[intersectorSize];
  interval *tmpXInts = (interval*)malloc(number_of_input_bboxes*sizeof(interval));
  interval *tmpYInts = (interval*)malloc(number_of_input_bboxes*sizeof(interval));
  
  for(int i = 0; i < number_of_input_bboxes; i++) {
	mbr cMBR = input_bboxes[i];
	tmpXInts[i] = interval(cMBR.getId(), cMBR.getMinX(), cMBR.getMaxX());
	tmpYInts[i] = interval(cMBR.getId(), cMBR.getMinY(), cMBR.getMaxY());
  }
  
    maximal_interval_stabbing_new *x_intervals = new maximal_interval_stabbing_new(tmpXInts, number_of_input_bboxes, atoi(argv[argc - 4]), atoi(argv[argc - 3]), argv[argc - 5], atoi(argv[argc - 2]));
    maximal_interval_stabbing_new *y_intervals = new maximal_interval_stabbing_new(tmpYInts, number_of_input_bboxes, atoi(argv[argc - 4]), atoi(argv[argc - 3]), argv[argc - 5], atoi(argv[argc - 2]));

  //printf("Estructuras creadas..\n");
  //printf("%u;\n", is.memoryUsage());
  
  
  long* resultWT = (long*) malloc(number_of_input_bboxes * sizeof(long));
  int k;
  for(k = 4; k < argc - 7; k++) {
    //read input queries
    uint number_of_input_queries;
    mbr *input_queries;
    if(queryFileType == 0) {
      mbr::readMBRs(argv[k], &input_queries, &number_of_input_queries);
    } else if(queryFileType == 1) {
      mbr::readMBRsAndAssignIDs(argv[k], &input_queries, &number_of_input_queries);
    } else {
      printf("Not supported query type.");
    }
    
    uint i,j;
    uint resultSizeWT;
    double time = 0;
    
    if(!strcmp(argv[argc - 1], "time")) {
      for(i = 0; i < number_of_input_queries; i++) {
	tTiempo mitiempo1;
	tTiempo mitiempo2;
	getTiempo(&mitiempo1);
	
	for(j = 0; j < 5; j++) {
	  resultSizeWT = 0;
	  query(&input_queries[i], resultWT, &resultSizeWT, intersector, intersectorSize, x_intervals, y_intervals);
	}
	
	getTiempo(&mitiempo2);
	time += ((mitiempo2.user - mitiempo1.user) / 5);
      }
      fprintf(outputFile, "%lf;", time / number_of_input_queries);
    } else if(!strcmp(argv[argc - 1], "test")) {
      uint error = 0;
      for(i = 0; i < number_of_input_queries; i++) {
	resultSizeWT = 0;
	query(&input_queries[i], resultWT, &resultSizeWT, intersector, intersectorSize, x_intervals, y_intervals);
	printf("Query %ld: %d elements: ", input_queries[i].getId(), (uint)resultSizeWT);
	qsort(resultWT, resultSizeWT, sizeof(long), cmplong);
	for(j = 0; j < resultSizeWT; j++) {
	  printf("%ld; ", resultWT[j]);
	}
	printf("\n");
	
      }
      if(!error) {
	fprintf(outputFile, "All tests OK.\n");
      }
    } else {
      printf("Unknown option: %s\n", argv[argc - 1]);
    }
    
    free(input_queries);
  }
  
  free(resultWT);
  free(input_bboxes);
  
  if(!strcmp(argv[argc - 1], "time")) {
    fprintf(outputFile, "%u;", x_intervals->memoryUsage()+y_intervals->memoryUsage());
  }
  fprintf(outputFile, "\n");
  fclose(outputFile);
  
  return 0;
}
