#include "mbr_range.h"

mbr_range::mbr_range(mbr *mbrs, uint size, uint type, uint lowerBlockSize, uint upperBlockSize, char *encoder, int scaleFactor) {
  interval *tmpXInts = (interval*)malloc(size*sizeof(interval));
  interval *tmpYInts = (interval*)malloc(size*sizeof(interval));
  
  for(int i = 0; i < size; i++) {
	mbr cMBR = mbrs[i];
	tmpXInts[i] = interval(cMBR.getId(), cMBR.getMinX(), cMBR.getMaxX());
	tmpYInts[i] = interval(cMBR.getId(), cMBR.getMinY(), cMBR.getMaxY());
  }

  switch(type) {
	case NO_COMPACT:
	  x_intervals = new isaac_interval_stabbing(tmpXInts, size);
	  y_intervals = new isaac_interval_stabbing(tmpYInts, size);
	  break;
	case COMPACT:
	  x_intervals = new compact_interval_stabbing(tmpXInts, size, lowerBlockSize, encoder, scaleFactor);
	  y_intervals = new compact_interval_stabbing(tmpYInts, size, lowerBlockSize, encoder, scaleFactor);
	  break;
	case COMPACT_NEW:
	  x_intervals = new compact_interval_stabbing_new(tmpXInts, size, lowerBlockSize, upperBlockSize, encoder, scaleFactor);
	  y_intervals = new compact_interval_stabbing_new(tmpYInts, size, lowerBlockSize, upperBlockSize, encoder, scaleFactor);
	  break;
	case MAXIMAL:
	  x_intervals = new maximal_interval_stabbing(tmpXInts, size, lowerBlockSize, upperBlockSize, encoder, scaleFactor);
	  y_intervals = new maximal_interval_stabbing(tmpYInts, size, lowerBlockSize, upperBlockSize, encoder, scaleFactor);
	  break;
  }
  
  free(tmpXInts);
  free(tmpYInts);
  
  intersectorSize = size/W+1;
  intersector = new uint[intersectorSize];
}

mbr_range::~mbr_range() {
  delete []intersector;
  delete x_intervals;
  delete y_intervals;
}

int mbr_range::query(mbr *queryMBR, long *result, uint *resultSize) {
  cleanIntersector();
  long id = queryMBR->getId();
  double minx = queryMBR->getMinX();
  double maxx = queryMBR->getMaxX();
  double miny = queryMBR->getMinY();
  double maxy = queryMBR->getMaxY();
  interval x_interval = interval(id, minx, maxx);
  interval y_interval = interval(id, miny, maxy);
  
  //x_intervals->query(&x_interval, result, resultSize, intersector, QUERY_TYPE_MARK);
  //y_intervals->query(&y_interval, result, resultSize, intersector, QUERY_TYPE_CHECK_REPORT);
  x_intervals->queryMark(&x_interval, result, resultSize, intersector);
  y_intervals->queryReport(&y_interval, result, resultSize, intersector);
}

uint mbr_range::memoryUsage() {
  return x_intervals->memoryUsage() + y_intervals->memoryUsage() + sizeof(mbr_range);
}